#lang cur

(require
 (only-in
  cur
  ; I'm going to rename these, because I'm mucking about with universes.
  [Type Type-]
  [~Type ~Type-]))

;; \gg ≫
;; \succ ≻
(define-typed-syntax Type
  [(_ s:layer n:nat) ≫
   #:with n+1 (+ (syntax-e #'n) 1)
   ---------------------
   [≻ #,(syntax-property
         (syntax-property (cur-expand #'(Type- n)) 'layer (syntax->datum #'s))
         ':
         #'(Type s n+1))]])

(begin-for-syntax
  (define-syntax-class layer
    ; either e (exceptional), m (mediation), or p (pure)
    (pattern (~or (~datum e) (~datum m) (~datum p)))))

(begin-for-syntax
  (require (for-syntax racket syntax/parse))
  (define-syntax ~Type
    (pattern-expander
     (syntax-parser
       [(_ s i)
        #'(~and (~and b (~Type- i))
                (~parse (~and s (~var _ layer))
                        (syntax-property #'b 'layer)))]))))

(begin-for-syntax
  (current-typecheck-relation
   (lambda (t1 t2)
     ; Since universes are type checked lazily, if the type we're checking (t1)
     ; is a universe, force it,
     (define t1+
       (syntax-parse t1
         [((~literal Type) _ _) ((current-type-eval) t1)]
         [((~literal Type-) _) ((current-type-eval) t1)]
         [_ t1]))
     (syntax-parse (list t1+ t2)
       [((~Type s1 n) (~Type s2 m))
        (and
         (equal? (syntax-e #'s1) (syntax-e #'s2))
         (<= (syntax-e #'n) (syntax-e #'m)))]
       [((~Type- n) (~Type (~datum p) m))
        (<= (syntax-e #'n) (syntax-e #'m))]
       [((~Type (~datum p) n) (~Type- m))
        (<= (syntax-e #'n) (syntax-e #'m))]
       [((~Π [x1 : τ_in1] τ_out1) (~Π [x2 : τ_in2] τ_out2))
        (and (cur-equal? #'τ_in1 #'τ_in2)
             (cur-type-check? (subst #'x2 #'x1 #'τ_out1) #'τ_out2))]
       [_ (cur-equal? t1+ t2)]))))

;;; Exceptions
;;; -----------------------------------------------------------
(define-type E : (Type e 0))

; String literals are exceptions

; Import string literals from Racket.
(require (only-in racket/base [#%datum #%datum-]))

; And give them a type.
(define-typed-syntax #%datum
  [(_ . s:str) ≫
   -----------
   [⊢ (#%datum- . s) ⇒ E]])

; Define the run-time representation of raise, which is Racket's raise, but it
; ignores the first argument (which will be a Cur type)
(module raise- racket/base
  (provide raise--)
  (define (raise-- _ e)
    (raise e)))

(require 'raise-)

(define-typerule/red (raise A e) ≫
  [⊢ A ≫ A- ⇒ (~Type s i)]
  #:fail-unless (equal? 'e (syntax-e #'s))
  (type-error #:src #'A #:msg "expected a type from the exceptional universes, but found ~a of universe (Type ~a ~a)" #'A #'s #'i)
  [⊢ e ≫ e- ⇐ E]
  ---------------------------------
  [⊢ (raise/eval A- e-) ⇒ A-]
  #:where raise/eval
  [(raise-- (~Π (x : A) B) M) ~> (λ (x : A) (raise B M))])

(define-type Bool-e : (Type e 0))
(define-type true-e : Bool-e)
(define-type false-e : Bool-e)

(define-typerule/red (rec-Bool-e v P m1 m2) ≫
  [⊢ v ≫ v- ⇐ Bool-e]
  [⊢ P ≫ P- ⇒ (~Π (x : ~Bool-e) (~Type (~datum e) i))]
  [⊢ m1 ≫ m1- ⇐ (P- true-e)]
  [⊢ m2 ≫ m2- ⇐ (P- false-e)]
  ---------------------
  [⊢ (rec-Bool-e/eval v- P- m1- m2-) ⇒ (P- v-)]
  #:where rec-Bool-e/eval
  [(#%plain-app ~true-e P m1 m2) ~> m1]
  [(#%plain-app ~false-e P m1 m2) ~> m2])

(define-type Bool : (Type p 0))
(define-type true : Bool)
(define-type false : Bool)

(define-syntax :type-of
  (lambda (stx)
    (syntax-parse stx
      [(_ e)
       (displayln (cur->datum (cur-type-infer #'e)))
       #'e])))

(module+ test
  (require rackunit rackunit/turnstile)
  ; Cumulativity working?
  (check-type Bool-e : (Type e 0))
  (check-type Bool-e : (Type e 1))
  (check-type Bool-e : (Type e 2))

  (check-type true-e : Bool-e)
  (check-type false-e : Bool-e)

  (check-type Bool : (Type p 0))
  (check-not-type Bool : (Type e 0))

  (check-type "Meow" : E)

  (check-type (raise Bool-e "meow") : Bool-e)

  (check-equal?
   (rec-Bool-e true-e (λ (x : Bool-e) Bool-e) false-e (raise Bool-e "Ooops"))
   false-e)

  (require (only-in racket/function [thunk thunk-]))
  (check-exn (lambda- (x) (equal? x "meow")) (thunk- (raise Bool-e "meow")))
  #;(check-not-type (raise Bool "meow") : Bool))
